#!/usr/bin/perl

#---------------------------------------------------------------------
# Copyright (C) 2003 Chad Redman <chad_eu2@katica.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#---------------------------------------------------------------------


# File:     EU2/Parse/GenericParser.pm
# Author:   Chad Redman <chad_eu2@katica.org>

# $Id: GenericParser.pm,v 1.4 2005/08/27 03:26:01 chad Exp $

package Symbol;

sub new {
    my ($pkg, $text, $line, $character) = @_;
    my $self = [$text, $line, $character];
    return bless $self, $pkg;
}

sub text { my $self = shift; return $self->[0]; }
sub line { my $self = shift; return $self->[1]; }
sub char { my $self = shift; return $self->[2]; }
sub location { my $self = shift; return (@$self[1,2]); }


package EU2::Parse::GenericParser;

$EU2::Parse::GenericParser::VERSION = '0.9.4';
$EU2::Parse::GenericParser::OPT_CHECK_RANGE = 'strict';
$EU2::Parse::GenericParser::OPT_CHECK_YEAR = '0:9999';   ## For stricter checking, change to 1419:1821
$EU2::Parse::GenericParser::ERRORFUNCTION = sub { print STDERR @_, "\n"; };

sub optCheckRange {
    my $self = shift;
    if (ref $self) {
        my $value = shift;
        if (defined $value) {
            return $self->{_OPTCHECKRANGE}=$value;
        } else {
            return $self->{_OPTCHECKRANGE};
        }
    } else {
        #my $value = $self;
        my $value = shift;
        if (defined $value) {
            return $OPT_CHECK_RANGE=$value;
        } else {
            return $OPT_CHECK_RANGE;
        }
    }
}

sub optCheckYear {
    my $self = shift;
    if (ref $self) {
        my $value = shift;
        if (defined $value) {
            return $self->{_OPTCHECKYEAR}=$value;
        } else {
            return $self->{_OPTCHECKYEAR};
        }
    } else {
        #my $value = $self;
        my $value = shift;
        if (defined $value) {
            return $OPT_CHECK_YEAR=$value;
        } else {
            return $OPT_CHECK_YEAR;
        }
    }
}

### Dates use text for month. This maps each to a number
%db_months = qw(
    january    01
    february   02
    march      03
    april      04
    may        05
    june       06
    july       07
    august     08
    september  09
    october    10
    november   11
    december   12
);


sub ErrorFunc(&) {
    my $pkg = shift;
    ($ERRORFUNCTION) = @_;
}

sub Log {
    my $pkg = shift;
    &$ERRORFUNCTION(@_);
}

sub Version {
    return $VERSION;
}

sub new {
    my ($pkg) = @_;
    my $self = {
        'symbols' => [],
        '_OPTCHECKRANGE' => $OPT_CHECK_RANGE,
        '_OPTCHECKYEAR' => $OPT_CHECK_YEAR,
    };

    return bless $self, $pkg;
}

sub fileHandle {
    my $self = shift;
    if (@_) { $self->{fileHandle} = shift }
    return $self->{fileHandle};
}

sub source {
    my $self = shift;
    if (@_) { $self->{source} = shift }
    return $self->{source};
}


sub addSymbol {
    my ($self, $sym) = @_;
    push @{$self->{symbols}}, $sym;
}


sub _tokenize {
    my $self = shift;

    my @lines = ();

    if ($self->fileHandle) {
        my $fh = $self->fileHandle;
        @lines = <$fh>;
        close $fh;
    } else {
        @lines = split(/\r?\n/, $self->source);
    }

    if (!@lines) {
        Log 0, "Error: No input for Event Parser!\n";
        return 0;
    }

    my @tokens = ();
    my $line;
    my $line_no = 0;
    my $char_no = 0;
    my $tmp_inquote = undef;
    while (@lines) {
        $line = splice(@lines,0,1);
        ++$line_no;
        $char_no = 0;
        while (length($line)) {
            if ($line =~ /^\s+/) { $char_no += length($&); $line = $'; next;}        # whitespace
            if ($line =~ /^\#.*/) { $char_no += length($&); $line = $'; next;}        #comment

            if ((!defined $tmp_inquote) && ($line =~ /^\"(.*)\"\s*$/)) { $self->addSymbol(new Symbol($1, $line_no, $char_no)); $char_no += length($&); $line = $'; next;}
                #quoted string #1. Matches "Zweite Armeecorps "Neumark" " (nested double-quotes, can only parse by assuming it takes up a single line)

            if ((!defined $tmp_inquote) && ($line =~ /^\"([^\"]*)\"/)) { $self->addSymbol(new Symbol($1, $line_no, $char_no)); $char_no += length($&); $line = $'; next;}
                #quoted string #2. basic--not nesting.

            if ((!defined $tmp_inquote) && ($line =~ /^\"([^\"\r\n]*)/)) { $tmp_inquote = $1; $char_no += length($&); $line = $'; next;}
                #Start of quoted string spread over multiple lines

            if ((defined $tmp_inquote) && ($line =~ /^([^\"]*)\"/)) { $self->addSymbol(new Symbol($tmp_inquote . ' ' . $1, $line_no, $char_no)); $char_no += length($&); $line = $'; $tmp_inquote = undef; next;}
                #Continued quoted string terminates on this line

            if ((defined $tmp_inquote) && ($line =~ /^([^\r\n]*)/)) { $tmp_inquote .= ' ' . $1; $char_no += length($&); $line = $'; next;}
                #Unterminated quoted string still not terminated

            if ($line =~ /^([0-9+-.]+)/) { $self->addSymbol(new Symbol($1, $line_no, $char_no)); $char_no += length($&); $line = $'; next;}
                #number
            if ($line =~ /^([A-Za-z0-9_.]+|=|\{|\})/) { $self->addSymbol(new Symbol($1, $line_no, $char_no)); $char_no += length($&); $line = $'; next;}
                #any other valid token
            else {
                Log 0, "Warning: unlikely token at line $line_no char $char_no : '$line'";
                if ($line =~ /^(\S+)/) { $self->addSymbol(new Symbol($1, $line_no, $char_no)); $char_no += length($&); $line = $'; next;}
                ## The next non-whitespace string of characters. Not a valid token, but save it anyway to keep the flow going
            }
        }
    }
}


#Virtual: parse($self)
#Virtual: tokenMatches($self, $match, $src)

#  ------- UTILITIES ---------
sub matches {            # (\@tokens, @fields)
    my ($self, $toks, @fields) = @_;     ## first arg is a package or object name

    foreach (0 .. $#fields) {
        return 0 if !defined $toks->[$_];

        my $match = $fields[$_];
        my $src = $toks->[$_]->text;

        ### First, try to match the token within the instantiated class. Result can be:
        ###   1 = match,  0 = error (e.g., invalid enum or out of range), -1 = undetermined (keep trying)
        my $inst_match = $self->tokenMatches($match, $src);
        if ($inst_match == 1) {
            next;
        } elsif ($inst_match == 0) {
            return 0;
        }
        # remaining option is -1, which means to try the other options within the current class

        ## A number can be 0, +/- 1...N, or +/- "inf". A "v" after a number indicates to ignore if the lenient option is on
        if ($match =~ /^:int,(-?[0-9]+|-inf)(v?),(-?[0-9]+|inf)(v?):$/) {
            my ($lo, $lo_v, $hi, $hi_v) = ($1, $2, $3, $4);

            return 0 if $src !~ /^(-?[0-9]+)$/;

            if ($self->optCheckRange ne 'none') {
                if (($lo ne '-inf') && ($src < $lo ) && !($lo_v && $self->optCheckRange eq 'loose')) {
                    Log 0, "Integer $src out of range ($lo to $hi)";
                    return 0;
                }

                if (($hi ne 'inf') && ($src > $hi ) && !($hi_v && $self->optCheckRange eq 'loose')) {
                    Log 0, "Integer $src out of range ($lo to $hi)";
                    return 0;
                }
            }
        } elsif ($match =~ /^:province(,rand)?:$/) {
            next if (($1 eq ',rand') and ($src <= -1 and $src >= -4));
            return 0 if $src !~ /^[0-9]+$/;
            if ($src < 1 or $src > 1614) {
                Log 0, "Integer $src out of range ($1 to $2)";
                return 0;
            }
        } elsif ($match eq ':year:') {
            return 0 if $src !~ /^-?[0-9]+$/;
            if ($self->optCheckYear) {
                my ($lo, $hi) = split(/[-,:]/, $self->optCheckYear);
                if ($src < $lo or $src > $hi) {
                    Log 0, "Year $src out of range ($lo to $hi)";
                    return 0;
                }
            }
        } elsif ($match eq ':month:') {
            if (!defined $db_months{lc $src}) {
                Log 0, "Month '$src' invalid";
                return 0;
            }
        } else {
            return 0 if $src !~ /^$match$/ and lc($src) !~ /^$match$/;
        }
   }

   1;
}


sub balanceParen {
    my ($toks, $pre, $post) = @_;
    my $count = 1;

    while (@$toks) {
        my $cur = (shift @$toks)->text;
        if ($cur eq $pre) { ++$count; }
        elsif ($cur eq $post) { --$count; return if $count == 0; }
    }

    Log 0, "Unmatched '$pre' before end of buffer after '" . join(' ', @$toks[0..10]) . "'";
    return;
}


sub warnBadState {          #($state, $buf);
    my ($class, $state, $toks) = @_;
    my @jointags;
    foreach ( 0 .. 10) {
        push @jointags, ($toks->[$_]->text) if defined $toks->[$_];
    }

    Log 0, "Syntax error in object $class, state $state, at line "
                   . $toks->[0]->line . " char "
                   . $toks->[0]->char . ", with buffer '" . join(' ', @jointags) . "'";
}

sub _abortParseError {
    my ($tokens) = @_;
    #EventParser->warnBadState('??', $tokens);

    #my $time = time();
    #open (ERRLOG, "> err_$time.log");       ## just ignore errors, instead of exiting
    #print ERRLOG join(' ', @$tokens) . "\n";
    #close ERRLOG;
    return 0;
}

sub formatDate {
    my ($year, $monthstr, $day) = @_;
    if (!defined $db_months{lc $monthstr}) {
        Log 0, "formatDate: Unknown month $monthstr";
        return sprintf("%04d/%02d/%02d", $year, 0, $day);
    } else {
        return sprintf("%04d/%02d/%02d", $year, $db_months{lc $monthstr}, $day);
    }
}


1;
