#!/usr/bin/perl

#---------------------------------------------------------------------
# Copyright (C) 2003 Chad Redman <chad_eu2@katica.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#---------------------------------------------------------------------


# File:     lint.pl
# Author:   Chad Redman <chad_eu2@katica.org>

# $Id: lint.pl,v 1.2 2003/07/22 04:44:36 chad Exp $

use FileHandle;
use Getopt::Long;
use EU2::Parse::LeaderParser;
use EU2::Parse::MonarchParser;
use EU2::Parse::EventParser;

$VERSION = '0.9.2';

EU2::Parse::GenericParser->ErrorFunc (sub {print STDOUT @_, "\n";});

if (@ARGV == 0) {
    &Syntax;
    exit 1;
}

GetOptions('help', 'version', 'check-range=s', 'check-year=s', 'nowarn-dup-ids',
           'leaders=s@', 'monarchs=s@', 'events=s@',
           'datadir=s',
           'importleaders:s', 'importmonarchs:s',
           'textcsv=s@', 'countrycsv=s', 'eventlist=s@', 'eventlistfilter=s@') or exit &Syntax();


if ($opt_help) {
    &CmdHelp;
    exit 0;
}

if ($opt_version) {
    print "This is lint.pl, version $VERSION, with parser engine ", EU2::Parse::EventParser->Version, "\n";
    exit 0;
}



sub CmdHelp {
    print STDOUT <<'EOT';
Command
-------
    perl lint.pl [options]

Or, if you are not running it from the current directory:
    perl -I <lint_directory> <lint_directory>\lint.pl [options]


Options
-------
-help
    Displays this message
-version
    Print the current version of this script, and the parser version it is
    using
-check-range <type>
    Determines how much to complain if event values are out of reasonable
    ranges. Type can be strict, loose, or none. Default is strict
-check-year <low:high>
    Warns if any date is out of this year range. Default is 0:9999
-nowarn-dup-ids
    Don't complain about duplicate IDs being used. Not fully tested
-leaders <file_or_directory> [-leaders file_or_directory ...]
    File or directory to read leaders from. It will ignore any file which
    doesn't match "leaders.TAG". If -countrycsv is specified, it will also
    ignore any country which isn't defined in country.csv
-monarchs <file_or_directory> [-monarchs file_or_directory ...]
    File(s) or directory(s) to read monarchs from. It will ignore any file
    which doesn't match "monarchs.TAG". If -countrycsv is specified, it
    will also ignore any country which isn't defined in country.csv
-countrycsv <path_to_country.csv>
    Location of the country.csv file. If defined, it will read in country
    tags, and use them to determine which leaders.* and monarchs.* to read
-importleaders
-importmonarchs
    Instead of getting leader/monarch IDs by parsing leaders.* or
    monarchs.* files, import them from the text files that get created in
    the lint_data directory from a previous parse. These two parameters are
    used when you want to cross-references leader IDs with monarch and
    event data, or to cross-reference both leaders and monarchs in event
    data. It will speed up testing by avoiding a re-check of files that
    have already been checked
-importleaders <filename> [-importleaders <filename> ...]
-importmonarchs <filename> [-importmonarchs <filename> ...]
    Same as above, but specify a specific filename, rather than the default
    lint_data\leaders.dat and lint_data\monachs.dat files. Multiple files
    can be specified.
-datadir <directory_name>
    Use a different directory to store program output. If not specified,
    all output will be written to the directory "lint_data" within the
    current directory. The directory will be created if it does not exist.
-events <file_or_directory> [-events <file_or_directory> ...]
    File(s) or directory(s) to read events from.
-textcsv <path_to_text.csv> [-textcsv <path_to_text.csv> ...]
    If specified, the program will open a text.csv file, and make note of
    any tag that looks like EVENT* or ACTION*. After this, when events are
    analyzed, it will verify that any name or desc reference that looks
    like EVENT* or ACTION* is actually defined in the text.csv file.
-eventlist <filename> [-eventlist <filename> ...]
    Reads in a scenario include file (e.g., events.txt), and searches for
    'event = "*"' patterns. It will then use all the files defined here as
    the basis for which of the event files/directories specified above to
    read or ignore; It will ignore any event file not defined in the list
    file. If not specified, all event files will be included, and none will
    be ignored.
-eventlistfilter <pattern> [-eventlistfilter <pattern> ...]
    This is used in combination with the -eventlist directive. The event
    files references in scenario include files usually include a partial
    path name such as 'Db\Events\filename.txt' instead of 'filename.txt'.
    If you specify pattern "Db\Events", then this will be stripped off. If
    not specified, any directory names will be filtered, leaving just the
    bare filename.
EOT

    exit 0;
}


sub Syntax {
    print STDERR <<'EOT';
Syntax: perl lint.pl [options]
Use the -help option for a list of all command line parameters
EOT
    exit 1;
}


######## Start of main program ########

{
    $opt_datadir = './lint_data' unless defined $opt_datadir;
    use File::Path('mkpath');
    mkpath($opt_datadir, 1);
}


EU2::Parse::LeaderParser->optCheckRange($opt_check_range) if defined $opt_check_range;
EU2::Parse::MonarchParser->optCheckRange($opt_check_range) if defined $opt_check_range;
EU2::Parse::EventParser->optCheckRange($opt_check_range) if defined $opt_check_range;

EU2::Parse::LeaderParser->optCheckYear($opt_check_year) if defined $opt_check_year;
EU2::Parse::MonarchParser->optCheckYear($opt_check_year) if defined $opt_check_year;
EU2::Parse::EventParser->optCheckYear($opt_check_year) if defined $opt_check_year;

my (%leaderdb, %monarchdb, %eventdb, %textdb, %countrydb, %eventlistdb);

if (defined @opt_textcsv) {
    %textdb = &readTextData( @opt_textcsv );
}

if (defined $opt_countrycsv) {
    %countrydb = &readTextData( $opt_countrycsv );
}

if (@opt_leaders) {
    my @leaderfiles = &getFiles( \@opt_leaders );
    %leaderdb = &readLeaders( \@leaderfiles );
    &dumpData("$opt_datadir/leaders.dat", \%leaderdb, [qw(ID FILE LINE NAME CATEGORY STARTDATE DEATHDATE RANK MOVEMENT FIRE SHOCK SIEGE SPECIAL DORMANT LOCATION REMARK)] );
}

if (@opt_monarchs) {
    my @monarchfiles = &getFiles( \@opt_monarchs );
    %monarchdb = &readMonarchs( \@monarchfiles, \%leaderdb );   ## Use the leaderdb to check for ID collisions
    &dumpData("$opt_datadir/monarchs.dat", \%monarchdb, [qw(ID FILE LINE NAME STARTDATE DEATHDATE DIP ADM MIL DORMANT REMARK)] );
}

if (defined $opt_importleaders) {
    %leaderdb = &restoreData($opt_importleaders || "$opt_datadir/leaders.dat");
}

if (defined $opt_importmonarchs) {
    %monarchdb = &restoreData($opt_importmonarchs || "$opt_datadir/monarchs.dat");
}

if (@opt_eventlist) {
    %eventlistdb = &readEventList( \@opt_eventlistfilter, \@opt_eventlist );
}

if (@opt_events) {
    my @eventfiles = &getFiles( \@opt_events );
    my %eventdb = &readEvents( \@eventfiles, \%leaderdb, \%monarchdb, \%textdb, \%eventlistdb );   ## Cross-check any leader or monarch references
                                                                          ## TODO: Cross-check string references
    &dumpData("$opt_datadir/events.dat", \%eventdb, [qw(ID FILE LINE NAME DESC MONARCH_REFS LEADER_REFS EVENT_REFS)] );
}


#print "Event ID ranges: " . join("\n", @db), "\n";
#
#if ($opt_full_id_report) {
#    my $first = '';
#    my $last = '';
#    foreach my $id (sort { $a <=>$b } keys %ids) {
#        if ($first eq '') {
#            $first = $id;
#        }
#
#        if ($last eq '') {
#            $last = $id;
#        } elsif ($id == $last+1) {
#            $last = $id;
#        } else {
#            push @idranges, (($first == $last) ? $first : "$first-$last");
#            $first = $id;
#            $last = '';
#        }
#    }
#
#    push @idranges, (($first == $last) ? $first : "$first-$last");
#
#    print "Event ID ranges: " . join(',', @idranges), "\n";
#
#
#}


## getFiles: Expand directory names to the files contained within
sub getFiles {
    my $names = shift;    # an arrayref containing either file or directory names
    my @filelist;

    foreach my $arg (@$names) {
        if (-d $arg) {
            if (opendir DIR, $arg) {
                push @filelist, sort(grep -T, map "$arg/$_", readdir DIR);           ## include all text files
                closedir DIR;
            } else {
                warn "Error opening directory $arg : $!\n";
            }
        } else {
            push @filelist, $arg;
        }
    }

    return @filelist;
}


### readLeaders: parse each file passed in. Return a hash of
### id: [id, linenum, name, startdate, ...] for all ids
sub readLeaders {
    my $filelist = shift;
    my %ids = ();

    foreach my $file (@$filelist) {
        if ($file =~ /\bleaders\.(.{3})$/i) {
            if (scalar %countrydb) {
                if (!defined $countrydb{uc $1}) {
                    print "    **** (Skipped: $file -- country " . uc($1) . " not defined) ****\n";
                    next;
                }
            }
        } else {
            print "    **** (Skipped: $file -- doesn't match 'Leaders.XXX') ****\n";
            next;
        }

        my $fh = new FileHandle("< $file");
        if (defined $fh) {
            my $parser = new EU2::Parse::LeaderParser();
            $parser->fileHandle($fh);
            print "    **** $file ****\n";
            $parser->parse;

            foreach my $leader (@{$parser->leaders}) {
                my $id = $leader->{ID};
                if (defined $ids{$id}) {
                    $opt_nowarn_dup_ids or $parser->Log(
                            "Duplicate Leader ID '$id', first defined in $ids{$id}->[1], Line $ids{$id}->[2]");
                } else {
                    $ids{$id} = [ $leader->{ID}, $file, $leader->{LINE}, $leader->{NAME}, $leader->{CATEGORY},
                              $leader->{STARTDATE},
                              $leader->{DEATHDATE},
                              $leader->{RANK},
                              $leader->{MOVEMENT},
                              $leader->{FIRE},
                              $leader->{SHOCK},
                              $leader->{SIEGE},
                              $leader->{SPECIAL},
                              $leader->{DORMANT},
                              $leader->{LOCATION},
                              $leader->{REMARK} ];
                }
            }
        } else {
            warn "!!! Error opening file $file: $!\n";
            next;
        }
    }

    return %ids;
}


### readMonarchs: parse each file passed in. Return a hash of
### id: [id, linenum, name, startdate, ...] for all ids
sub readMonarchs {
    my ($filelist, $hleaders) = @_;
    my %ids = ();

    foreach my $file (@$filelist) {
        if ($file =~ /\bmonarchs\.(.{3})$/i) {
            if (scalar %countrydb) {
                if (!defined $countrydb{uc $1}) {
                    print "    **** (Skipped: $file -- country " . uc($1) . " not defined) ****\n";
                    next;
                }
            }
        } else {
            print "    **** (Skipped: $file -- doesn't match 'Monarchs.XXX') ****\n";
            next;
        }

        my $fh = new FileHandle("< $file");
        if (defined $fh) {
            my $parser = new EU2::Parse::MonarchParser();
            $parser->fileHandle($fh);
            print "    **** $file ****\n";
            $parser->parse;

            foreach my $monarch (@{$parser->monarchs}) {
                my $id = $monarch->{ID};
                if (defined $ids{$id}) {
                    $opt_nowarn_dup_ids or $parser->Log(
                            "Duplicate Monarch ID '$id', first defined in $ids{$id}->[1], Line $ids{$id}->[2]");
                } elsif (defined $hleaders->{$id}) {
                    $opt_nowarn_dup_ids or $parser->Log(
                            "Duplicate Monarch ID (already used as a leader ID) '$id', first " .
                            "defined in $hleaders->{$id}->[1], Line $hleaders->{$id}->[2]");
                } else {
                    $ids{$id} = [ $monarch->{ID}, $file, $monarch->{LINE}, $monarch->{NAME}, $monarch->{STARTDATE},
                              $monarch->{DEATHDATE}, $monarch->{DIP}, $monarch->{ADM}, $monarch->{MIL},
                              $monarch->{DORMANT}, $monarch->{REMARK} ];
                }

            }
        } else {
            print STDERR "!!! Error opening file $file: $!\n";
            next;
        }
    }

    return %ids;
}


### readEvents: parse each file passed in. Return a hash of
### id: [id, linenum, name, startdate, ...] for all ids. Cross-check references leaders and monarch IDs
sub readEvents {
    my ($filelist, $hleaders, $hmonarchs, $htext, $heventlist) = @_;
    my %ids = ();

    foreach my $file (@$filelist) {
        if (scalar %$heventlist) {
            my ($basename) = ($file =~ /[\\\/]([\w .]+)$/);   ## Get the base filename
            if (!defined $heventlist->{uc $basename}) {
                print "    **** (Skipped: $file -- Not defined in eventlist) ****\n";
                next;
            }
        } else {
            if ($file =~ /(Johan\.rst|eventlist\.txt)$/i) {
                print "    **** (Skipped: $file -- matches special ignore list) ****\n";
                next;
            }
        }
        my $fh = new FileHandle("< $file");
        if (defined $fh) {
            my $parser = new EU2::Parse::EventParser();
            $parser->fileHandle($fh);
            print "    **** $file ****\n";
            $parser->parse;

            foreach my $event (@{$parser->events}) {
                my $id = $event->{ID};
                if (defined $ids{$id}) {
                    $opt_nowarn_dup_ids or $parser->Log(
                            "Duplicate Event ID '$id', first defined in $ids{$id}->[1], Line $ids{$id}->[2]");
                } else {
                    $ids{$id} = [ $event->{ID},
                                  $file,
                                  $event->{LINE},
                                  $event->{NAME},
                                  $event->{DESC},
                                  $event->{MONARCH_REFS},
                                  $event->{LEADER_REFS},
                                  $event->{EVENT_REFS}, ];
                }

                ## Check refs to leader and monarch IDs. Save event ID ref checking until all events have been parsed
                if (scalar %$hleaders) {
                    foreach my $refid ( @{$event->{LEADER_REFS}} ) {
                        if (!defined $hleaders->{$refid}) {
                            $parser->Log("Undefined leader ID '$refid', in event $id (Line $ids{$id}->[2])");
                        }
                        if (defined $hmonarchs->{$refid}) {
                            $parser->Log("Leader reference '$refid' is actually a monarch ID, in event $id (Line $ids{$id}->[2])");
                        }
                    }
                }
                if (scalar %$hmonarchs) {
                    foreach my $refid ( @{$event->{MONARCH_REFS}} ) {
                        if (!defined $hmonarchs->{$refid}) {
                            $parser->Log("Undefined monarch ID '$refid', in event $id (Line $ids{$id}->[2])");
                        }
                        if (defined $hleaders->{$refid}) {
                            $parser->Log("Monarch reference '$refid' is actually a leader ID, in event $id (Line $ids{$id}->[2])");
                        }
                    }
                }

                ## Check string refs that look like EVENTNAME, EVENTHIST, or ACTIONNAME
                if (scalar %$htext) {
                    foreach ( $event->{NAME}, $event->{DESC}, @{$event->{ACTIONNAME_REFS}} ) {
                        next unless (/^(EVENT|ACTIONNAME)\w+/);
                        if (!defined $htext->{$_}) {
                            $parser->Log("Invalid text reference '$_', in event $id (Line $ids{$id}->[2])");
                        }
                    }
                }
            }
        } else {
            warn "!!! Error opening file $file: $!\n";
            next;
        }
    }

    ## Now that the event ids have been populated, validate cross-refs to them
    ## Sort the found events first by filename, then by line number
    foreach my $evdata ( sort{$a->[1] cmp $b->[1] or $a->[2] <=> $b->[2]} values %ids ) {
        foreach my $refid ( sort @{$evdata->[7]} ) {
            if (!defined $ids{$refid}) {
                EU2::Parse::EventParser->Log("Undefined reference to event ID '$refid', in event $evdata->[0] (File $evdata->[1], Line $evdata->[2])");
            }

            if ($refid == $evdata->[0]) {
                EU2::Parse::EventParser->Log("Command action in event '$evdata->[0]' refers to itself (File $evdata->[1], Line " . $evdata->[2] . ")");
            }
        }
    }

    return %ids;
}


### dumpData: Write out arrayrefs to a tab-delimited file
sub dumpData {
    my ($file, $rhash, $header) = @_;

    open(FILE, "> $file") or warn "Error opening file '$file' for writing : $!\n";
    $header and print FILE join("\t", @$header), "\n";

    foreach my $row ( values %$rhash ) {
        print FILE join("\t",  map { (ref $_ eq 'ARRAY') ? join(',', @$_) : $_ } @$row), "\n";
    }
    close FILE;
}


### restoreData: Create a hash out of tab-separated data. Assume first field is the key (but also include with values)
sub restoreData {
    my ($file) = @_;
    my @res;

    open(FILE, "< $file") or warn "Could not open file '$file' : $!\n";
    while (<FILE>) {
        chomp;
        my @fields = split("\t", $_);
        push @res, $fields[0], [ @fields ];
    }
    close FILE;
    return @res
}


### readTextData: Read in the text key IDs and make a key=>1 hash from it (caller doesn't care about the value)
###
sub readTextData {
    my (@files) = @_;
    my @res;

    foreach my $file (@files) {
        open(FILE, "< $file") or warn "Could not open file '$file' : $!\n";
        while (<FILE>) {
            #next unless (/^(EVENT|ACTION)/);
            chomp;
            my @fields = split(";", $_);
            push @res, $fields[0], 1;
        }
    }
    close FILE;
    return @res
}

### readEventList: Use a scenario file to define which event files are in use. The $dirfilters parameter is
### used to filter out the pathname
###
sub readEventList {
    my ($dirfilters, $files) = @_;

    my @res;

    foreach my $file ( @$files ) {
        open(FILE, "< $file") or warn "Could not open file '$file' : $!\n";
        while (<FILE>) {
            next unless /^event\s*=\s*\"([^\"]+)\"/;
            my $tmp = $1;

            if (@$dirfilters) {
                foreach $pat ( map {quotemeta} @$dirfilters ) {
                    if ($tmp =~ /^$pat[\\\/]?/i) {
                        $tmp = $';
                        next;
                    }
                }
            } else {
                #No specific eventlist filter is specified. Assume they want the basename in all cases
                $tmp =~ s/.*[\\\/]([\w .]+)$/$1/;   ## Get the base filename
            }

            push @res, uc($tmp), $file;
            print "DEBUG: Found reference to $tmp in event list $file\n";
        }
    }

    close FILE;
    return @res;
}
